###############################################################################
#
#  Copyright (C) 2007-TODAY OpenERP SA. All Rights Reserved.
#
#  $Id$
#
#  Developed by OpenERP (http://openerp.com) and Axelor (http://axelor.com).
#
#  The OpenERP web client is distributed under the "OpenERP Public License".
#  It's based on Mozilla Public License Version (MPL) 1.1 with following 
#  restrictions:
#
#  -   All names, links and logos of OpenERP must be kept as in original
#      distribution without any changes in all software screens, especially
#      in start-up page and the software header, even if the application
#      source code has been changed or updated or code has been added.
#
#  You can see the MPL licence at: http://www.mozilla.org/MPL/MPL-1.1.html
#
###############################################################################

import cherrypy

from openobject.widgets import Widget
from openobject.widgets import InputWidget
import simplejson

__all__ = ['TinyWidget', 'TinyInputWidget', 'ConcurrencyInfo',
           'register_widget', 'get_widget', 'get_registered_widgets',
           'InputWidgetLabel']

_attrs_boolean = {
    'select': False,
    'nolabel': False,
    'required': False,
    'readonly': False,
}

def _boolean_attr(attrs, name):

    if name not in _attrs_boolean:
        return attrs.get(name)

    val = attrs.get(name)
    if isinstance(val, basestring) and val.lower() in ('false', 'none', '0'):
        return False

    return (attrs.get(name) and True) or _attrs_boolean.get(name)

class TinyWidget(Widget):

    params = [
        'colspan',
        'rowspan',
        'string',
        'nolabel',
        'visible',
        'valign',
        'model',
        'label',
        'editable_style'
    ]

    colspan = 1
    rowspan = 1
    string = None
    nolabel = False
    visible = True
    model = None

    valign = "middle"

    def __init__(self, **attrs):

        super(TinyWidget, self).__init__(**attrs)

        prefix = attrs.get('prefix', '')
        self._name = prefix + (prefix and '/' or '') + attrs.get('name', '')

        self.colspan = int(self.colspan)
        self.rowspan = int(self.rowspan)
        self.nolabel = _boolean_attr(attrs, 'nolabel')

        self.visible = True
        self.editable_style = attrs.get('editable_style', False) and '1' or False
        try:
            visval = attrs.get('invisible', 'False')
            ctx = attrs.get('context', {})
            self.invisible = eval(visval, {'context': ctx})
        except:
            pass

        self.attributes = attrs.get('attrs', {})

    def get_widgets_by_name(self, name, kind=Widget, parent=None):

        result = []
        parent = parent or self

        for wid in parent.iter_member_widgets():

            if wid.name == name and isinstance(wid, kind):
                result.append(wid)

            if getattr(wid, 'member_widgets', False):
                result += self.get_widgets_by_name(name, kind=kind, parent=wid)

        return result

    @property
    def name(self):
        return self._name

class InputWidgetLabel(Widget):
    template = "/openerp/widgets/templates/label.mako"
    params = ['string', 'help']

    def __init__(self, name, string, help=None):
        super(InputWidgetLabel, self).__init__(name=name, string=string, help=help)

class TinyInputWidget(TinyWidget, InputWidget):
    params = [
        'select',
        'required',
        'readonly',
        'force_readonly',
        'help',
        'editable',
        'translatable',
        'inline',
        'states',
        'callback',
        'change_default',
        'onchange',
        'kind',
        'hide_default_menu',
        'filters', # filter buttons within an input widget, part of the same implicit "group"
        'readonly_before_state',
        'required_error_msg'
    ]

    select = False
    required = False
    readonly = False
    force_readonly = False
    help = None
    editable = True
    translatable = False
    inline = False

    states = None
    callback = None
    change_default = None
    kind = None
    hide_default_menu = False
    required_error_msg = None

    label_type = InputWidgetLabel

    def __init__(self, **attrs):

        super(TinyInputWidget, self).__init__(**attrs)

        if isinstance(self.states, basestring):
            self.states = self.states.split(',')

        self.select = _boolean_attr(attrs, 'select')
        self.required = _boolean_attr(attrs, 'required')
        self.readonly = _boolean_attr(attrs, 'readonly')
        self.force_readonly = _boolean_attr(attrs, 'force_readonly')

        self.translatable = attrs.get('translate', False)

        self.readonly_before_state = self.readonly
        self.set_state(attrs.get('state', 'draft'))

        self.callback = attrs.get('on_change', None)
        self.kind = attrs.get('type', None)
        self.hide_default_menu = _boolean_attr(attrs, 'hide_default_menu')

        self.default_focus = attrs.get('default_focus', False)
        self.required_error_msg = attrs.get('required_error_msg', False)
        self.label = self.label_type(self.name, self.string, self.help)
        self.filters = []

    def set_state(self, state):
        if isinstance(self.states, dict) and state in self.states:
            attrs = dict(self.states[state])

            self.readonly = attrs.get('readonly', self.readonly)
            self.required = attrs.get('required', self.required)
            self.default = attrs.get('value', self.default)

    def get_value(self):
        """Get the value of the field.

        @return: field value
        """
        return self.default

    def set_value(self, value):
        """Set the value of the field.

        @param value: the value
        """
        if isinstance(value, basestring):
            value = ustr(value)

        self.default = value

    def get_display_value(self):
        """Get the display value of the field.
        """

        try:
            return self.validator.from_python(self.default)
        except:
            pass

        return self.get_value()

    def update_params(self, params):
        super(TinyInputWidget, self).update_params(params)

        params.update(
            kind=self.kind,
            editable=self.editable,
            inline=self.inline,
            hide_default_menu=self.hide_default_menu,
            attrs={
                #'attrs': self.attributes or None,
                'change_default': self.change_default or None,
                'callback': self.callback or None,
                'onchange': self.onchange
            })

        if self.required_error_msg:
            params['attrs']['required_error_msg'] = self.required_error_msg
        if self.hide_default_menu:
            params['attrs']['hide_default_menu'] = '1'

        if self.default_focus:
            params['attrs']['autofocus'] = 'autofocus'
        if self.readonly:
            params['attrs']['disabled'] = 'disabled'
        if self.editable_style:
             params['attrs']['editable_style'] = '1'
        #if self.states:
        #    states = self.states
        #    # convert into JS
        #    if isinstance(states, dict):
        #        states = dict([(k, dict(v)) for k, v in states.iteritems()])
        #    params['attrs']['states'] = simplejson.dumps(states)


class ConcurrencyInfo(TinyInputWidget):
    template = "/openerp/widgets/templates/concurrencyinfo.mako"

    params = ['ids', 'model', 'info']

    def __init__(self, model, ids):
        super(ConcurrencyInfo, self).__init__(model=model, ids=ids)

    @property
    def info(self):
        return getattr(cherrypy.request, 'terp_concurrency_info', {})
    @classmethod
    def update(cls, resource, records):
        info = getattr(cherrypy.request, 'terp_concurrency_info', {})
        vals = info.setdefault(resource, {})
        for item in records:
            vals[item['id']] = item.pop('__last_update', '')
        cherrypy.request.terp_concurrency_info = info
        cherrypy.response.headers['X-Concurrency-Info'] = \
            simplejson.dumps(info)


from openobject import pooler

def register_widget(klass, types, view="form"):
    """Register a widget class for the given view and types

    @param view: the view type (e.g. form, tree)
    @param types: register for the give types
    @param klass: widget class
    """

    if not isinstance(types, (list, tuple)):
        types = [types]

    for t in types:
        pooler.register_object(klass, key=t, group=view)


def get_widget(type, view="form"):
    """Get the widget of the given type for the given view.

    @param view: the view
    @param type: the widget type
    """

    pool = pooler.get_pool()
    return pool.get(type, group=view)


def get_registered_widgets(view="form"):
    """Get all the registered widgets for the given view type.

    @param view: the view
    @returns: dict of all the registered widgets
    """
    pool = pooler.get_pool()
    return pool.get_group(view)


# vim: ts=4 sts=4 sw=4 si et
